import torch
from torch.utils.data import DataLoader
from diresa_torch.arch.models import Diresa
from diresa_torch.utils.utils import _r2_score, _set_components_to_mean


def encode_diresa(
    model: Diresa,
    data_loader: DataLoader,
    input_filter: callable = lambda data: data,
) -> torch.Tensor:
    """
    Encodes a dataset into a latent representation.
    Prerequisite: latent components are already ordered.

    :param model: model to use
    :param data_loader: data to be encoded
    :param input_filter: function used to filter input data, default is no filtering
    :return: latent representation of the dataset
    """
    assert model.is_ordered(), "Model must be ordered"
    device = next(model.parameters()).device
    latent = []

    model.eval()
    with torch.no_grad():
        for data in data_loader:
            data = data.to(device)
            input_data = input_filter(data)
            outputs = model.encode(input_data)
            latent.append(outputs.cpu())

    return torch.cat(latent, dim=0)


def latent_r2_per_variable(
    model: Diresa,
    data_loader: DataLoader,
    incr: bool = False,
    input_filter: callable = lambda data: data,
    target_filter: callable = lambda data: data,
    verbose: bool = False,
) -> list:
    """
    Computes R2 scores of latent components per variable. Variables should be on axis 1 (first is 0).
    If this is not the case, input_filter and target_filter can be used to swap axes.
    Prerequisite: latent components are already ordered.

    :param model: Diresa model
    :param data_loader: DataLoader
    :param incr: If True incremental R2 score are calculated, default is False
    :param input_filter: Function used to filter input data, default is no filtering
    :param target_filter: Function used to filter target data, default is no filtering
    :param verbose: If True, prints most important component per variable, default is False
    :return: R2 scores of latent components per variable, shape (latent_dim, nbr of variables)
    """
    assert model.is_ordered(), "Model must be ordered"
    device = next(model.parameters()).device
    num_batches = len(data_loader)
    r2 = 0

    model.eval()
    with torch.no_grad():
        for _, data in enumerate(data_loader):
            data = data.to(device)
            input_data = input_filter(data)
            target_data = target_filter(data)
            latent = model.encode(input_data)  # (batch_size, latent_size)
            latent_size = latent.shape[1]
            nbr_vars = target_data.shape[1]
            r2_batch = torch.zeros(latent_size, nbr_vars, device=device)

            # For each latent dimension i:
            #   mask others -> decode -> compute R2 per field
            for i in range(latent_size):
                if not incr:
                    latent_masked = _set_components_to_mean(latent, i)  # (batch_size, latent_size)
                else:
                    latent_masked = _set_components_to_mean(latent, list(range(0, i+1)))
                # decode
                pred = model.decode(latent_masked)
                # compute R2 per output field j
                for j in range(nbr_vars):
                    r2_batch[i, j] = _r2_score(target_data[:, j], pred[:, j])
            r2 += r2_batch
        r2 /= num_batches  # shape (latent_size, nbr_vars)

    if incr:
        for i in range(latent_size - 1, 0, -1):
            r2[i] = r2[i] - r2[i - 1]

    if verbose:
        if incr:
            print("Latent R² scores are incremental!")
        for i in range(latent_size):
            # best field index for latent dim i
            j_best = torch.argmax(r2[i]).item()
            score = r2[i, j_best].item()
            print(f"Latent dimension {i:02d} is most useful for reproducing "
                  f"{j_best} (R² = {score:.4f})")
        for j in range(nbr_vars):
            i_best = torch.argmax(r2[:, j]).item()
            score = r2[i_best, j].item()
            print(f"Field {j:02d} is best explained by latent dimension {i_best:02d} (R² = {score:.4f})")

    return r2


def latent_vectors(
    model: Diresa,
    data_loader: DataLoader,
    factor: float = 0.5,
    incr: bool = False,
    input_filter: callable = lambda data: data,
) -> torch.Tensor:
    """
    Calculates decoded latent vectors. Prerequisite: model must be ordered.
    See: https://journals.ametsoc.org/view/journals/aies/4/3/AIES-D-24-0034.1.xml
    Appendix D: c.Latent variable interpretation.

    :param model: Diresa model
    :param data_loader: DataLoader
    :param factor: Multiplication factor for  standard deviation
    :param incr: If True incremental vectors are calculated, default is False
    :param input_filter: Function used to filter input data, default is no filtering
    :return: Decoded latent vectors
    """
    assert model.is_ordered(), "Model must be ordered"
    device = next(model.parameters()).device
    num_batches = len(data_loader)
    mean_values = 0.
    std_dev = 0.

    model.eval()
    with torch.no_grad():
        for _, data in enumerate(data_loader):
            data = data.to(device)
            input_data = input_filter(data)
            latent = model.encode(input_data)  # (batch_size, latent_size)

            mean_values += torch.mean(latent, dim=0)
            std_dev += torch.std(latent, dim=0)

        mean_values /= num_batches
        std_dev /= num_batches

        latent_dim = mean_values.shape[0]
        latent_plus = mean_values.repeat(latent_dim, 1)  # (latent_size, latent_size)
        latent_minus = mean_values.repeat(latent_dim, 1)

        if not incr:
            for i in range(latent_dim):
                latent_plus[i, i] += factor * std_dev[i]
                latent_minus[i, i] -= factor * std_dev[i]
        else:
            for i in range(latent_dim):
                latent_plus[i, :i+1] += factor * std_dev[:i+1]
                latent_minus[i, :i+1] -= factor * std_dev[:i+1]

        decoded_plus = model.decode(latent_plus)  # (latent_size, ...)
        decoded_minus = model.decode(latent_minus)
        vectors = decoded_plus - decoded_minus  # (latent_size, ...)

        if incr:
            for i in range(latent_dim - 1, 0, -1):
                vectors[i] = vectors[i] - vectors[i-1]

    return vectors



import torch
import torch.nn as nn

from diresa_torch.arch.modules import DenseLayer, DistanceLayer, CNNLayer, OrderingLayer
from diresa_torch.arch.config import DiresaSetup, DiresaConfig
from typing import Tuple, Optional
from functools import reduce, wraps
import operator

import logging


def _cnn_output_flat(nb_stack, dims, cnn_output_filter):
    """
    Computes intermediate dimensions from CNN to FNN.
    Output dimensions of CNN is `(original_x / 2^len(stack)) * (original_y / 2^len(stack)) * output filter`.
    :param nb_stack: number of stacks in CNN models. Len of nb_stack defines how many times the input dimensions are divided by 2 due to the MaxPool2d layer.
    :param dims: original input dimensions
    :param cnn_output_filter: filters of the last CNN layer.
    """
    spatial_dims = [dim // (2**nb_stack) for dim in dims]
    return reduce(operator.mul, spatial_dims + [cnn_output_filter])


class _Encoder(nn.Module):
    """
    Provides the encoder for DIRESA.

    :param config: `DiresaConfig` object built by the static builder method DiresaSetup.create_config()
    """

    def __init__(self, config: DiresaConfig):
        super().__init__()

        if config.dense_config and not config.cnn_config:
            # NOTE:
            # Input of dense_units is input_shape[0].
            # By passing input dimension (config.input_shape[0])
            # by copy here we do not modify the original input_shape
            # If we mutate config.input_shape here in place we need to remember to NOT
            # modify it a second time in the decoder otherwise 2 layers are added.
            dense_units = (config.input_shape[0],) + tuple(config.dense_config.dense_units)
            self.network = DenseLayer(dense_units, config.activation)

            # This removes the last activation layer as DenseLayer always puts an activation function
            # after each Layer
            del self.network[-1]

        elif config.cnn_config and not config.dense_config:

            self.cnn_encoder = CNNLayer(
                config.cnn_config.stack,
                config.cnn_config.stack_filters,
                config.cnn_config.kernel_size,
                config.input_shape,
                config.activation,
            )

            # NOTE: If no dense_config is specified, there should be no activation function in the last layer.
            # In last block (-1) we remove the last RELU (-2). We do not change this in CNNLayer object as it
            # would require to pass in information about the structure of DIRESA into the CNNLayer class.
            # Fusing CNN and FFN structures should be done in the Encoder/Decoder models and not in submodules.
            del self.cnn_encoder[-1][-2]  # -2 as there is a MaxPool2D at [-1]

            self.flatten = nn.Flatten()

            self.network = nn.Sequential(
                self.cnn_encoder,
                self.flatten,
            )

        elif config.cnn_config and config.dense_config:

            self.cnn_encoder = CNNLayer(
                config.cnn_config.stack,
                config.cnn_config.stack_filters,
                config.cnn_config.kernel_size,
                config.input_shape,
                config.activation,
            )

            ffn_input = _cnn_output_flat(
                len(config.cnn_config.stack),
                config.input_shape[1:],
                config.cnn_config.stack_filters[-1],
            )

            dense_units = (ffn_input,) + tuple(config.dense_config.dense_units)

            self.ffn_encoder = DenseLayer(dense_units, config.activation)

            # need to delete last layer which is the activation function
            del self.ffn_encoder[-1]

            self.flatten = nn.Flatten()

            self.network = nn.Sequential(
                self.cnn_encoder, self.flatten, self.ffn_encoder
            )

    def __getitem__(self, key):
        """Enable subscriptable access"""
        return self.network[key]

    def __delitem__(self, key):
        """Enable item deletion"""
        del self.network[key]

    def forward(self, x):
        return self.network(x)


class _Decoder(nn.Module):
    def __init__(self, config: DiresaConfig):
        super().__init__()

        if config.dense_config and not config.cnn_config:
            # NOTE: Do not change config.input_shape in place, see comment in Encoder
            # Case where we only have dense units. Input of dense_units is input_shape[0]
            # (3, 64, 64, 2) where 3 is input and 2 is latent space will become
            # (64, 64, 3) when reversing.
            dense_units = (config.input_shape[0],) + tuple(config.dense_config.dense_units)

            self.network = DenseLayer(dense_units, config.activation, reverse=True)

            # this deletes the last activation layer as DenseLayer always puts an activation function
            # after each Linear layer.
            del self.network[-1]

        elif config.cnn_config and not config.dense_config:
            # spatial dimensions at the end of the CNN layer
            spatial_dims = [
                dim // (2 ** len(config.cnn_config.stack))
                for dim in config.input_shape[1:]
            ]

            # flatten output from the cnn which depends on dimensions of x/y and the final number
            # of filters.
            cnn_flat_dims = _cnn_output_flat(
                len(config.cnn_config.stack),
                config.input_shape[1:],
                config.cnn_config.stack_filters[-1],
            )

            unflatten_shape = (
                cnn_flat_dims // (spatial_dims[0] * spatial_dims[1]),
                *spatial_dims,
            )

            # need to unflatten from latent space to correct cnn input dimension
            self.unflatten = nn.Unflatten(1, unflatten_shape)

            self.cnn_decoder = CNNLayer(
                config.cnn_config.stack,
                config.cnn_config.stack_filters,
                config.cnn_config.kernel_size,
                config.input_shape,
                config.activation,
                reverse=True,
            )

            # Activation function is at the last position in the decoder
            del self.cnn_decoder[-1][-1]

            self.network = nn.Sequential(self.unflatten, self.cnn_decoder)

        elif config.cnn_config and config.dense_config:
            ffn_input = _cnn_output_flat(
                len(config.cnn_config.stack),
                config.input_shape[1:],
                config.cnn_config.stack_filters[-1],
            )

            # NOTE:
            # Will be reversed when building dense layers such that
            # last layer of fully connected feeding into CNN has the right (flatten) output
            # dimension.
            dense_units = (ffn_input,) + tuple(config.dense_config.dense_units)

            self.ffn_decoder = DenseLayer(dense_units, config.activation, reverse=True)

            spatial_dims = [
                dim // (2 ** len(config.cnn_config.stack))
                for dim in config.input_shape[1:]
            ]

            unflatten_shape = (config.cnn_config.stack_filters[-1], *spatial_dims)

            self.unflatten = nn.Unflatten(1, unflatten_shape)

            self.cnn_decoder = CNNLayer(
                config.cnn_config.stack,
                config.cnn_config.stack_filters,
                config.cnn_config.kernel_size,
                config.input_shape,
                config.activation,
                reverse=True,
            )
            # Delete activation function which is last element in last block
            del self.cnn_decoder[-1][-1]

            self.network = nn.Sequential(
                self.ffn_decoder, self.unflatten, self.cnn_decoder
            )

    def __getitem__(self, key):
        """Enable subscriptable access"""
        return self.network[key]

    def __delitem__(self, key):
        """Enable item deletion"""
        del self.network[key]

    def forward(self, x):
        return self.network(x)


class Diresa(nn.Module):
    """
    Distance Regularized Autoencoder.

    Can be built using class method ``from_hyper_param`` which builds the encoder and decoder from a list of hyperparameters.
    DIRESA can also be built by providing a custom encoder and decoder using class method ``from_custom``.
    """

    @classmethod
    def from_hyper_param(
        cls,
        # Global parameters
        input_shape: Tuple[int, ...] = (),
        activation: nn.Module = nn.ReLU(),
        # Dense parameters
        dense_units: Optional[Tuple[int, ...]] = None,
        # CNN parameters
        stack: Optional[tuple[int, ...]] = None,
        stack_filters: Optional[Tuple[int, ...]] = None,
        kernel_size: Tuple[int, int] = (3, 3),
    ):
        """
        Builds DIRESA from hyperparameters.
        DIRESA can be built using fully connected layers, CNN layers or a combination of both.

        The argument `dense_units` is used to add fully connected layers, where the length of the tuple represents depth of
        the network and each individual value represents the number of nodes per layer. First element describes the first
        hidden layer. Last layer defines the output layer which is the size of latent space. E.g. ``dense_units = (16, 32, 64)``
        is a fully connected network with two hidden layers `(16, 32)` and last layer which defines latent space of size `64`.
        When only one value is present it represents a linear projection on to a latent space with dimensions of that value
        (as for the last layer no activation is used).

        To build a convolutional encoder, parameters `stack` and `stack_filters` have to be provided. `stack` handles the number
        of convolutional layers per block and `stack_filters` handles the number of filters per block. That is, if
        ``stack = (2, 1)`` and ``stack_filters = (16, 8)``, the encoder network is going to be split as
        `[Conv2D 16 -> Activation -> Conv2D 16 -> Activation] -> MaxPool -> [Conv2D 8] -> MaxPool -> Flatten`.
        Each block is always followed by a `MaxPool` layer and at the end a `Flatten` layer is added.

        If dense parameters and convolutional parameters are both specified the network is built as follows:
        `[CNN] -> [Dense] -> LatentSpace -> [Dense] -> [CNN]`.


        :param input_shape: is the shape of input features
        :param activation: activation function used by all layers, except for the last layer of the encoder/decoder, which doesn't have an activation function
        :param stack: the number of Conv2D layers in each block
        :param stack_filters: number of filters (out channels) for the Conv2D layers in each block, `len(stack_filters)` has always to be equal to `len(stack)`
        :param kernel_size: kernel size for all Conv2D layers
        :param dense_units: tuple which describes the width of each fully connected layer
        """

        # Create global configuration with sub-configs
        # General config is not saved at the moment in DIRESA class.
        # => Not possible to retrieve it by calling Diresa.config.
        config = DiresaSetup.create_config(
            stack=stack,
            stack_filters=stack_filters,
            kernel_size=kernel_size,
            input_shape=input_shape,
            dense_units=dense_units,
            activation=activation,
        )

        return cls(_Encoder(config), _Decoder(config))

    @classmethod
    def from_custom(cls, encoder: nn.Module, decoder: nn.Module):
        """
        Builds Diresa from custom provided encoder and decoder. Distance and
        ordering layers are added through the Diresa class.

        :param encoder: encoder model
        :param decoder: decoder model
        """
        return cls(encoder, decoder)

    # Could make constructor private to avoid direct instantiation.
    # Does not make a difference at the moment as there are
    # no additional checks in `from_custom`.

    def __init__(self, encoder: nn.Module, decoder: nn.Module):
        super().__init__()

        # Called base as those are the simple encoders/decoders which
        # produce latent space and decode latent space
        # but do not provide any additional features such as distance computation
        # or ordering
        self.base_encoder = encoder
        self.dist_layer: DistanceLayer = DistanceLayer(dim_less=True)
        self.base_decoder = decoder
        self.ordering_layer = OrderingLayer()

    def __random_ordering_index(self, x: torch.Tensor) -> torch.Tensor:
        """
        Shuffles batch x.
        """
        batch_size = x.shape[0]
        perm_indices = torch.randperm(batch_size, device=x.device)
        return perm_indices

    def get_ordering(self) -> Optional[torch.Tensor]:
        """
        :return: Tensor representing ordering of the latent components by `R2-score` or `None` if no ordering has been computed yet.
        """
        return self.ordering_layer.order

    def is_ordered(self) -> bool:
        """
        Returns `True` if ordering of the latent components has been computed on the model, `False` otherwise.

        :return: Ordering status
        """
        return self.ordering_layer.order is not None

    def reset_ordering(self) -> Optional[torch.Tensor]:
        """
        Removes current ordering and returns it.

        :return: Tensor representing current ordering or `None` if no ordering has been computed yet.
        """
        current = self.ordering_layer.order
        self.ordering_layer.order = None
        return current

    @staticmethod
    def __warn_ordering(f):
        """
        Decorator used to produce a warning when no ordering has been computed
        """

        # used to display docs from wrapped function
        @wraps(f)
        def check_order(*args, **kwargs):
            self = args[0]
            if not self.is_ordered():
                # NOTE: Could improve doc here
                logging.warning(
                    "Latent components of Diresa have not yet been ordered. To produce an ordering of latent components call order_diresa() from trainer module"
                )
            return f(*args, **kwargs)

        return check_order

    @__warn_ordering
    def encode(self, x: torch.Tensor) -> torch.Tensor:
        """
        Runs encoder to provide latent space representation, latent components are ordered by `R2-score` if ordering has been previously computed.

        :param x: batch of original data
        :return: encoded batch, ordered if ordering has been previously computed.
        """
        encoded = self.base_encoder(x)
        ordered = self.ordering_layer(encoded)
        return ordered

    @__warn_ordering
    def decode(self, x: torch.Tensor) -> torch.Tensor:
        """
        Runs decoder to provide reconstructed data from latent space representation (keeping ordering of the latent components in mind if ordering has been computed).

        :param x: batch of latent data.
        :return: decoded batch
        """
        reversed_ordering = self.ordering_layer(x, reverse=True)
        return self.base_decoder(reversed_ordering)

    # TODO: Could find a a naming scheme here to show
    # that this is part of the DIRESA arch.
    def _encode_with_distance(
        self,
        x: torch.Tensor,
        x_twin: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Encodes batch `x` into latent space and compute distance between encoder
        and twin encoder in original and latent space.

        :param x: Batch to encode
        :param x_twin: Twin batch to encode, if None then x_twin = shuffled(x)
        :return: (latent, distance) which are the encoded latent variables as well as
            the distance between the encoder and twin_encoder in original and latent space.
        """
        y: torch.Tensor = self.base_encoder(x)

        # Twin encoder
        if x_twin is None:
            # shuffle per batch
            index = self.__random_ordering_index(x)
            x_twin = x[index]
            y_twin = y[index]
        else:
            # user shuffling
            y_twin = self.base_encoder(x_twin)

        dist: torch.Tensor = self.dist_layer(x, x_twin, y, y_twin)
        return y, dist

    def forward(
        self,
        x: torch.Tensor,
        x_twin: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Produces ``reconstructed``, ``latent`` and ``distance`` information
        """
        if x_twin is None:
            latent, dist = self._encode_with_distance(x)
        else:
            latent, dist = self._encode_with_distance(x, x_twin)

        reconstructed = self.base_decoder(latent)

        return reconstructed, latent, dist

    def fast_eval(self, x: torch.Tensor) -> torch.Tensor:
        """
        Produces a reconstructed output by using only encoder and decoder.
        This skips the distance layer and ordering layer.

        :param x: batch to evaluate.
        """
        return self.base_decoder(self.base_encoder(x))



import torch
import torch.nn as nn
from enum import Enum

import logging


# NOTE:
# With current approach, input layer to DenseLayer is computed from either input_shape when no CNN is present or from the output of CNN. Adding this layer is the responsibility of the DIRESA class.


class ModuleList(Enum):
    """
    Enum class used for tests to select for which kind
    of module we are counting different layers.
    See test_structure.py
    """

    DenseLayer = 0
    CNNLayer = 1


class DenseLayer(nn.Module):
    """
    Provides Network with Feed Forward Fully Connected Layers.

    :param dense_units: tuple which describes the width of each feed forward layer. First element describes the first hidden layer. Last layer defines the output layer which is the size of latent space. E.g. ``(16, 32, 64)`` is a FFN with two hidden layers ``(16, 32)`` and last layer which defines latent space of size ``64``.
    :param activation: activation function. An activation function will be placed after each Linear Layer
    :param reverse: Builds opposite network specified by ``dense_units``. Does not take into account last value of ``dense_unit`` representing latent space, instead output layer is defined having size ``input_size``
    """

    def __init__(self, dense_units=(), activation=nn.ReLU(), reverse=False):
        super().__init__()

        if reverse:
            dense_units = dense_units[::-1]

        layers = []

        for i in range(len(dense_units) - 1):
            layers.append(nn.Linear(dense_units[i], dense_units[i + 1]))
            layers.append(activation)

        self.network = nn.Sequential(*layers)

    def forward(self, x):
        return self.network(x)

    def __getitem__(self, key):
        """Enable subscriptable access"""
        return self.network[key]

    def __delitem__(self, key):
        """Enable item deletion"""
        del self.network[key]


class DistanceLayer(nn.Module):
    """
    :param x1: batch of input samples to encoder
    :param x2: batch of input shuffled samples to twin encoder
    :param y1: batch of latent representations of encoder
    :param y2: batch of latent representations of twin encoder
    :return: batch of distances between inputs, batch of distances between latent representations
    """

    def __init__(self, dim_less=False):
        """
        :param dim_less: if True distance is divided by dimension of space
        """
        super().__init__()
        self.dim_less = dim_less

    def forward(self, x1, x2, y1, y2):
        """
        :param x1: batch of input samples to encoder
        :param x2: batch of input shuffled samples to twin encoder
        :param y1: batch of latent representations of encoder
        :param y2: batch of latent representations of twin encoder
        :return: batch of distances between inputs, batch of distances between latent representations
        """

        # Do not take the sqrt to produce euclidan distance, leads to better training
        dist1 = torch.square(x1 - x2)
        dist2 = torch.square(y1 - y2)
        dist1 = torch.sum(
            torch.reshape(dist1, [dist1.shape[0], -1]), dim=1
        )  # sum over all dims, except 0
        dist2 = torch.sum(
            torch.reshape(dist2, [dist2.shape[0], -1]), dim=1
        )  # sum over all dims, except 0

        if self.dim_less:
            # divide by size of 1 sample
            dist1 = torch.div(dist1, x1.numel() / x1.size(dim=0))
            dist2 = torch.div(dist2, y1.numel() / y1.size(dim=0))

        return torch.stack((dist1, dist2), dim=-1)


class _CNNBlock(nn.Module):
    """
    Builds CNN Block
    :param num_layers: number of layers in this block
    :param input: input channels
    :param output: output channels
    :param kernel_size: size of kernel used for the convolution operation
    :param activation: non linear activation function
    """

    def __init__(self, num_layers, input, output, kernel_size, activation=nn.ReLU()):
        super().__init__()

        layers = []

        # Maintain spatial dimensions
        padding = (kernel_size[0] // 2, kernel_size[1] // 2)

        for _ in range(num_layers):
            layers.append(nn.Conv2d(input, output, kernel_size, padding=padding))
            layers.append(activation)
            # change after 1st iteration so that we can stack multiple blocks
            input = output

        layers.append(nn.MaxPool2d((2, 2)))
        self.block = nn.Sequential(*layers)

    def forward(self, x):
        return self.block(x)

    def __getitem__(self, key):
        """Enable subscriptable access"""
        return self.block[key]

    def __delitem__(self, key):
        """Enable item deletion"""
        del self.block[key]


class _CNNTransposeBlock(nn.Module):
    """
    CNN Transpose Block for decoder - mirrors the encoder block but with transpose convolutions

    :param num_layers: number of layers in this block
    :param input: input channels
    :param output: output channels
    :param kernel_size: size of kernel used for the convolution operation
    :param activation: non linear activation function
    """

    def __init__(self, num_layers, input, output, kernel_size, activation=nn.ReLU()):
        super().__init__()
        layers = []

        # Need to reverse MaxPool2d
        layers.append(nn.Upsample(scale_factor=2, mode="nearest"))

        # Now build the transpose conv layers (in reverse order)
        current_channels = input
        for i in range(num_layers):
            # Calculate padding to maintain spatial dimensions
            padding = (kernel_size[0] // 2, kernel_size[1] // 2)

            # For the last layer, change to output channels
            if i == num_layers - 1:
                out_channels = output
            else:
                out_channels = current_channels

            layers.append(
                nn.ConvTranspose2d(
                    current_channels, out_channels, kernel_size, padding=padding
                )
            )
            layers.append(activation)
            current_channels = out_channels

        self.block = nn.Sequential(*layers)

    def forward(self, x):
        return self.block(x)

    def __getitem__(self, key):
        """Enable subscriptable access"""
        return self.block[key]

    def __delitem__(self, key):
        """Enable item deletion"""
        del self.block[key]


class CNNLayer(nn.Module):
    """
    Builds CNN layers for autoencoder.

    :param stack: are the number of blocks per layer
    :param stack_filters: are the number of output channels of each block (as input channels are defined by previous layer)
    :param kernel_size: size of kernel used for the convolution operation
    :param input_shape: shape of input data in following order (channels, x, y)
    :param activation: non linear activation function
    """

    def __init__(
        self,
        stack,
        stack_filters,
        kernel_size,
        input_shape,
        activation=nn.ReLU(),
        reverse=False,
    ):
        super().__init__()

        layers = []

        if reverse:
            # Decoder mode: reverse everything
            reversed_stack = stack[::-1]
            reversed_filters = stack_filters[::-1]

            # For decoder, we need to map back to original input channels
            # Input channels for decoder are the reversed filters
            input_channels = reversed_filters

            # input_shape is now output shape
            output_channels = tuple(reversed_filters[1:]) + (input_shape[0],)

            for block, input_ch, output_ch in zip(
                reversed_stack, input_channels, output_channels
            ):
                layers.append(
                    _CNNTransposeBlock(
                        block, input_ch, output_ch, kernel_size, activation
                    )
                )
        else:
            # Encoder Mode
            # renaming to fit with pytorch Docs
            output_channels = stack_filters

            # first dim of input_shape is depth (= number of channels), which are inputs of first CNN Block
            input_channels = tuple((input_shape[0],) + tuple(stack_filters)[:-1])

            for block, input_ch, output_ch in zip(
                stack, input_channels, output_channels
            ):
                layers.append(
                    _CNNBlock(block, input_ch, output_ch, kernel_size, activation)
                )

        self.cnn_layer = nn.Sequential(*layers)

    def forward(self, x):
        return self.cnn_layer(x)

    def __getitem__(self, key):
        """Enable subscriptable access"""
        return self.cnn_layer[key]

    def __delitem__(self, key):
        """Enable item deletion"""
        del self.cnn_layer[key]


class OrderingLayer(nn.Module):
    def __init__(self):
        super().__init__()

        # NOTE: The dimensions of the ordering are not known statically
        # as it is not possible to retrieve latent dimensions of an encoder
        # a user provides. Otherwise we could just a tensor.range(latent.shape)
        # to get an identity ordering (i.e. the ordering layer just passes through
        # the components)

        # Need to register the ordering as a buffer in order to move it to the correct device.
        # And save it to the state in order to get access to loading/
        self.register_buffer("_order", None)

    # Assigning and Reading self.order call those
    # methods internally. Could be used to return
    # an identity ordering in case self.order = None
    @property
    def order(self) -> torch.Tensor | None:
        return self._order

    @order.setter
    def order(self, order):
        if order is None:
            logging.info(f"Ordering has been set to {order}")
            self._order: torch.Tensor | None = None
        else:
            assert isinstance(
                order, torch.Tensor
            ), "Ordering should be provided as tensor"

            # Ordering tensor should be moved to the right device in
            # the ordering function. OrderingLayer does not statically have
            # access to models parameters and therefore to which device those
            # parameters are stored.
            self._order = order

    def forward(self, x: torch.Tensor, reverse=False):
        if self.order is not None:

            # NOTE: No idea if asserting within `forward` has a huge perfomance penalty
            assert (
                x.device == self.order.device
            ), "Ordering tensor and batch tensor should be located on the same device"

            if reverse:
                index = torch.argsort(self.order)
            else:
                index = self.order

            return torch.index_select(x, 1, index)

        # ignore ordering and just pass through
        else:
            return x
